#!/usr/bin/env python3
# coding: utf-8
"""
.. module:: BuscoAnalysis
   :synopsis: BuscoAnalysis implements general BUSCO analysis specifics
.. versionadded:: 3.0.0
.. versionchanged:: 5.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.
"""

from abc import ABCMeta, abstractmethod
from busco.BuscoConfig import BuscoConfig, BuscoConfigAuto
from busco.busco_tools.hmmer import HMMERRunner
import os
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log

logger = BuscoLogger.get_logger(__name__)


class BuscoAnalysis(metaclass=ABCMeta):
    """
    This abstract base class (ABC) defines methods required for most of BUSCO analyses and has to be extended
    by each specific analysis class
    """

    config = None

    def __init__(self):
        """
        1) load parameters
        2) load and validate tools
        3) check data and dataset integrity
        4) Ready for analysis
        """
        super().__init__()

        # Get paths
        self._lineage_results_dir = self.config.get("busco_run", "lineage_results_dir")
        self.main_out = self.config.get("busco_run", "main_out")
        self._working_dir = (
            os.path.join(self.main_out, "auto_lineage")
            if isinstance(self.config, BuscoConfigAuto)
            else self.main_out
        )
        self.run_folder = os.path.join(self._working_dir, self._lineage_results_dir)
        self._log_folder = os.path.join(self.main_out, "logs")

        # Get other useful variables
        self.input_file = self.config.get("busco_run", "in")
        self._lineage_dataset = self.config.get("busco_run", "lineage_dataset")
        self._lineage_name = os.path.basename(self._lineage_dataset)
        self._domain = self.config.get("busco_run", "domain")
        self._has_variants_file = os.path.exists(
            os.path.join(self._lineage_dataset, "ancestral_variants")
        )
        self._dataset_creation_date = self.config.get("busco_run", "creation_date")
        self.restart = self.config.getboolean("busco_run", "restart")

        self.gene_details = (
            {}
        )  # Dictionary containing coordinate information for predicted genes.
        self.headers = set()

        self._lineages_download_path = os.path.join(
            self.config.get("busco_run", "download_path"), "lineages"
        )

        self.hmmer_runner = None

        # Create optimized command line call for the given input
        # self.busco_type = "main" if isinstance(self._config, BuscoConfigMain) else "auto"
        # if self.busco_type == "main":
        #     self.set_rerun_busco_command(self._config.clargs)  # todo: rework rerun command

    @abstractmethod
    def cleanup(self):
        # Delete any non-decompressed files in busco_downloads
        try:
            for dataset_name in os.listdir(self._lineages_download_path):
                if dataset_name.endswith((".gz", ".tar")):
                    os.remove(dataset_name)
        except OSError:
            pass

    @abstractmethod
    @log(
        "Running BUSCO using lineage dataset {0} ({1}, {2})",
        logger,
        attr_name=["_lineage_name", "_domain", "_dataset_creation_date"],
        on_func_exit=True,
    )
    def run_analysis(self):
        """
        Abstract method, override to call all needed steps for running the child analysis.
        """
        self._create_dirs()
        self.init_tools()
        self._check_data_integrity()

    @log("***** Run HMMER on gene sequences *****", logger)
    def run_hmmer(self, input_sequences, busco_ids=None):
        """
        This function runs hmmsearch.
        """
        if not busco_ids:
            files = sorted(os.listdir(os.path.join(self._lineage_dataset, "hmms")))
            busco_ids = [
                os.path.splitext(f)[0] for f in files
            ]  # Each Busco ID has a HMM file of the form "<busco_id>.hmm"
        self.hmmer_runner.configure_runner(
            input_sequences, busco_ids, self._mode, self.gene_details
        )
        if self.restart and self.hmmer_runner.check_previous_completed_run():
            logger.info("Skipping HMMER run as output already processed")
        elif len(os.listdir(self.hmmer_runner.results_dir)) > 0:
            raise SystemExit(
                "HMMER results directory not empty. If you are running in restart mode, make sure you are "
                "using the same eukaryotic gene predictor (metaeuk/augustus) as before."
            )
        else:
            self.restart = False
            self.config.set("busco_run", "restart", str(self.restart))
            self.hmmer_runner.run()
        self.hmmer_runner.process_output()
        self.validate_output()
        self.hmmer_runner.filter()
        self.hmmer_runner.consolidate_busco_lists()
        output = self.hmmer_runner.create_output_content()
        self.hmmer_runner.write_hmmer_results(output)
        self.hmmer_runner.produce_hmmer_summary()
        return

    def validate_output(
        self,
    ):  # Transparent method that can be overwritten by child classes
        return

    @log("Checking dataset for HMM profiles", logger, debug=True)
    def _check_dataset_integrity(self):
        """
        Check the input dataset for hmm profiles, both files and folder are available
        Note: score and length cutoffs are checked when read by hmmer_runner: see _load_scores and _load_lengths
        Note: dataset.cfg file is not mandatory for offline mode
        # todo: implement a check for dataset.cfg file if not using offline mode

        :raises SystemExit: if the dataset is missing files or folders
        """

        # Check hmm files exist
        files = os.listdir(os.path.join(self._lineage_dataset, "hmms"))
        if not files:
            raise SystemExit(
                "The dataset you provided lacks hmm profiles in {}".format(
                    os.path.join(self._lineage_dataset, "hmms")
                )
            )

        if self._domain == "eukaryota":
            # Check prfl folder exists and contains profiles
            for dirpath, dirnames, files in os.walk(
                os.path.join(self._lineage_dataset, "prfl")
            ):
                if not files:
                    raise SystemExit(
                        "The dataset you provided lacks elements in {}".format(
                            os.path.join(self._lineage_dataset, "prfl")
                        )
                    )

        if not self._has_variants_file:
            logger.warning(
                "The dataset you provided does not contain the file ancestral_variants, likely because it "
                'is an old version. All blast steps will use the file "ancestral" instead'
            )

        return

    def _check_data_integrity(self):
        self._check_dataset_integrity()
        if not os.stat(self.input_file).st_size > 0:
            raise SystemExit("Input file is empty.")
        with open(self.input_file) as f:
            for line in f:
                if line.startswith(">"):
                    self._check_fasta_header(line)
                    self._check_seq_uniqueness(line)
        return

    def _check_seq_uniqueness(self, line):
        seq_id = line.split(" ")[0]
        if seq_id in self.headers:
            raise SystemExit("Duplicate of sequence {} in input file".format(seq_id))
        self.headers.add(seq_id)
        return

    @staticmethod
    def _check_fasta_header(header):
        """
        This function checks problematic characters in fasta headers,
        and warns the user and stops the execution
        :param header: a fasta header to check
        :type header: str
        :raises SystemExit: if a problematic character is found
        """
        for char in BuscoConfig.FORBIDDEN_HEADER_CHARS:
            if char in header:
                raise SystemExit(
                    'The character "%s" is present in the fasta header %s, '
                    "which will crash BUSCO. Please clean the header of your "
                    "input file." % (char, header.strip())
                )

        for char in BuscoConfig.FORBIDDEN_HEADER_CHARS_BEFORE_SPLIT:
            if char in header.split()[0]:
                raise SystemExit(
                    'The character "%s" is present in the fasta header %s, '
                    "which will crash Reader. Please clean the header of your"
                    " input file." % (char, header.split()[0].strip())
                )

        if header.split()[0] == ">":
            raise SystemExit(
                "A space is present in the fasta header %s, directly after "
                '">" which will crash Reader. Please clean the header of '
                "your input file." % (header.strip())
            )

    def _create_dirs(self):
        """
        Create the run (main) directory, log directory and the temporary directories
        :return:
        """
        self._create_main_dir()
        self._create_log_dir()
        # self._create_tmp_dir()

    def _create_log_dir(self):
        """
        Create a subfolder of the main output folder that contains all log files from BUSCO and the external tools used.
        :return:
        """
        if not os.path.exists(self._log_folder):
            os.mkdir(self._log_folder)
        return

    def _create_main_dir(self):
        """
        This function creates the run (main) directory
        :raises SystemExit: if write permissions are not available to the specified location
        """
        try:
            os.makedirs(self.run_folder)
        except FileExistsError:
            if not self.restart:
                raise SystemExit(
                    "Something went wrong. BUSCO stopped before overwriting run folder "
                    "{}".format(self.run_folder)
                )
        except PermissionError:
            raise SystemExit(
                "Cannot write to the output directory, please make sure "
                "you have write permissions to {}".format(self.run_folder)
            )
        return

    @log("Check all required tools are accessible...", logger, debug=True)
    def init_tools(self):
        """
        Init the tools needed for the analysis. HMMER is needed for all BUSCO analysis types.
        """
        self.hmmer_runner = HMMERRunner()
        return

    @property
    @abstractmethod
    def _mode(self):
        pass

    # def _run_tarzip_hmmer_output(self):  # todo: rewrite using tarfile
    #     """
    #     This function tarzips "hmmer_output" results folder
    #     """
    #     self._p_open(["tar", "-C", "%s" % self.run_folder, "-zcf", "%shmmer_output.tar.gz" % self.run_folder,
    #                   "hmmer_output", "--remove-files"], "bash", shell=False)
    #
    # @log("To reproduce this run: {}", logger, attr_name="_rerun_cmd", on_func_exit=True)
    # def set_rerun_busco_command(self, clargs):  # todo: reconfigure
    #     """
    #     This function sets the command line to call to reproduce this run
    #     """
    #
    #     # Find python script path
    #     entry_point = ""
    #     frame_ind = -1
    #     while "run_BUSCO.py" not in entry_point:
    #         entry_point = inspect.stack()[frame_ind].filename
    #         frame_ind -= 1
    #
    #     # Add required parameters and other options
    #     self._rerun_cmd = "python %s -i %s -o %s -l %s -m %s -c %s" % (entry_point, self._input_file, os.path.basename(self.main_out),
    #                                                                    self._lineage_dataset, self._mode, self._cpus)
    #
    #     try:
    #         if self._long:
    #             self._rerun_cmd += " --long"
    #         if self._region_limit != BuscoConfig.DEFAULT_ARGS_VALUES["limit"]:
    #             self._rerun_cmd += " --limit %s" % self._region_limit
    #         # if self._tmp != BuscoConfig.DEFAULT_ARGS_VALUES["tmp_path"]:
    #         #     self._rerun_cmd += " -t %s" % self._tmp
    #         if self._ev_cutoff != BuscoConfig.DEFAULT_ARGS_VALUES["evalue"]:
    #             self._rerun_cmd += " -e %s" % self._ev_cutoff
    #         # if self._tarzip:
    #         #     self._rerun_cmd += " -z"
    #     except AttributeError:
    #         pass
    #
    #     # Include any command line arguments issued by the user
    #     # arg_aliases = {"-i": "--in", "-o": "--out", "-l": "--lineage_dataset", "-m": "--mode", "-c": "--cpu",
    #     #                "-e": "--evalue", "-f": "--force", "-sp": "--species", "-z": "--tarzip",
    #     #                "-r": "--restart", "-q": "--quiet", "-v": "--version", "-h": "--help"}
    #     arg_aliases.update(dict(zip(arg_aliases.values(), arg_aliases.keys())))
    #     for a, arg in enumerate(clargs):
    #         if arg.startswith("-") and not arg in self._rerun_cmd:
    #             if arg in arg_aliases:
    #                 if arg_aliases[arg] in self._rerun_cmd:
    #                     continue
    #             if a + 1 < len(clargs) and not clargs[a + 1].startswith("-"):
    #                 self._rerun_cmd += " %s %s" % (arg, clargs[a + 1])
    #             else:
    #                 self._rerun_cmd += " %s" % arg
    #     return

    # TODO: catch unicode encoding exception and report invalid character line instead of doing content validation
    # todo: check config file exists before parsing



from busco.busco_tools.base import BaseRunner, NoGenesError
import os
import re
from collections import defaultdict
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from Bio import SeqIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
import shutil
import numpy as np
from configparser import NoOptionError
import subprocess

logger = BuscoLogger.get_logger(__name__)


class AugustusParsingError(Exception):
    def __init__(self):
        pass


class AugustusRunner(BaseRunner):

    ACCEPTED_PARAMETERS = [
        "strand",
        "genemodel",
        "singlestrand",
        "hintsfile",
        "extrinsicCfgFile",
        "maxDNAPieceSize",
        "protein",
        "introns",
        "start",
        "stop",
        "cds",
        "AUGUSTUS_CONFIG_PATH",
        "alternatives-from-evidence",
        "alternatives-from-sampling",
        "sample",
        "minexonintronprob",
        "minmeanexonintronprob",
        "maxtracks",
        "gff3",
        "UTR",
        "outfile",
        "noInFrameStop",
        "noprediction",
        "contentmodels",
        "translation_table",
        "temperature",
        "proteinprofile",
        "progress",
        "predictionStart",
        "predictionEnd",
        "uniqueGeneId",
    ]

    name = "augustus"
    cmd = "augustus"

    def __init__(self):
        self.gene_details = None
        self._augustus_config_path = os.environ.get("AUGUSTUS_CONFIG_PATH")
        self.config.set("busco_run", "augustus_config_path", self._augustus_config_path)
        self._target_species = self.config.get("busco_run", "augustus_species")
        super().__init__()
        self._output_folder = os.path.join(self.run_folder, "augustus_output")
        self.tmp_dir = os.path.join(self._output_folder, "tmp")
        self.pred_genes_dir_initial = os.path.join(
            self._output_folder, "predicted_genes_initial_run"
        )
        self.pred_genes_dir_rerun = os.path.join(
            self._output_folder, "predicted_genes_rerun"
        )
        self.extracted_prot_dir = os.path.join(
            self._output_folder, "extracted_proteins"
        )
        self.err_logfile = os.path.join(self.log_folder, "augustus_err.log")

        try:
            self.extra_params = self.config.get(
                "busco_run", "augustus_parameters"
            ).replace(",", " ")
        except NoOptionError:
            self.extra_params = ""
        self.chunksize = 10

        self.gff_dir = os.path.join(self._output_folder, "gff")
        self.err_logfiles = []
        self.any_gene_found = False
        self.param_keys = []
        self.param_values = []

        self.create_dirs([self._output_folder, self.extracted_prot_dir, self.gff_dir])

        self.init_checkpoint_file()

    def configure_runner(self, seqs_path, coords, sequences_aa, sequences_nt):
        self.run_number += 1

        # Placed here to allow reconfiguration for rerun
        self._target_species = self.config.get("busco_run", "augustus_species")

        self.check_tool_dependencies()
        self.gene_details = defaultdict(list)
        self.output_sequences = []

        self.seqs_path = seqs_path
        self.coords = coords

        self.sequences_aa = sequences_aa
        self.sequences_nt = sequences_nt

        self.pred_genes_dir = (
            self.pred_genes_dir_rerun
            if self.run_number == 2
            else self.pred_genes_dir_initial
        )

        # self.tmp_dir placed here to allow it to be recreated during reconfiguration for rerun
        self.create_dirs([self.pred_genes_dir, self.tmp_dir])

    @property
    def output_folder(self):
        return self._output_folder

    def check_tool_dependencies(self):
        """
        check dependencies on files and folders
        properly configured.
        :raises SystemExit: if Augustus config path is not writable or
        not set at all
        :raises SystemExit: if Augustus config path does not contain
        the needed species
        present
        """
        try:
            augustus_species_dir = os.path.join(self._augustus_config_path, "species")
            if not os.access(augustus_species_dir, os.W_OK):
                raise SystemExit(
                    "Cannot write to Augustus species folder, please make sure you have write "
                    "permissions to {}".format(augustus_species_dir)
                )

        except TypeError:
            raise SystemExit("The environment variable AUGUSTUS_CONFIG_PATH is not set")

        if not os.path.exists(os.path.join(augustus_species_dir, self._target_species)):
            # Exclude the case where this is a restarted run and the retraining parameters have already been moved.
            if (
                self.config.getboolean("busco_run", "restart")
                and self.run_number == 2
                and os.path.exists(
                    os.path.join(
                        self._output_folder,
                        "retraining_parameters",
                        self._target_species,
                    )
                )
            ):
                pass
            else:
                raise SystemExit(
                    'Impossible to locate the species "{0}" in Augustus species folder'
                    " ({1}), check that AUGUSTUS_CONFIG_PATH is properly set"
                    " and contains this species. \n\t\tSee the help if you want "
                    "to provide an alternative species".format(
                        self._target_species, augustus_species_dir
                    )
                )

    @log(
        "Running Augustus prediction using {} as species:",
        logger,
        attr_name="_target_species",
    )
    def run(self):
        super().run()
        if self.extra_params:
            logger.info(
                "Additional parameters for Augustus are {}: ".format(self.extra_params)
            )
            self.param_keys, self.param_values = self.parse_parameters()

        self.total = self._count_jobs()
        self.run_jobs()

    def process_output(self):
        logger.info("Extracting predicted proteins...")
        files = [
            f
            for f in sorted(os.listdir(self.pred_genes_dir))
            if any(busco_id in f for busco_id in self.coords)
        ]
        for filename in files:
            self._extract_genes_from_augustus_output(filename)

        if not self.any_gene_found and self.run_number == 1:
            raise NoGenesError("Augustus")

        self.gene_details = dict(self.gene_details)

        self._merge_stderr_logs()
        self._remove_individual_err_logs()

        return

    def _count_jobs(self):
        n = 0
        for busco_group, contigs in self.coords.items():
            for _ in contigs:
                n += 1
        return n

    def sort_jobs(self):
        jobs_size_info = []
        for busco_group, contigs in self.coords.items():

            for contig_name, contig_info in contigs.items():
                contig_start = contig_info["contig_start"]
                contig_end = contig_info["contig_end"]
                pred_size = int(contig_end) - int(contig_start)
                jobs_size_info.append(
                    {
                        "busco_group": busco_group,
                        "contig_name": contig_name,
                        "contig_start": contig_start,
                        "contig_end": contig_end,
                        "pred_size": pred_size,
                    }
                )
        job_sizes = [item["pred_size"] for item in jobs_size_info]
        new_job_order = np.argsort(job_sizes)[::-1]
        ordered_jobs = [jobs_size_info[i] for i in new_job_order]
        return ordered_jobs

    def generate_job_args(self):
        contig_ordinal_inds = defaultdict(int)
        njobs = 0

        ordered_jobs = self.sort_jobs()

        for job_info in ordered_jobs:
            contig_name = job_info["contig_name"]
            busco_group = job_info["busco_group"]
            contig_start = job_info["contig_start"]
            contig_end = job_info["contig_end"]
            contig_tmp_file = "{}.temp".format(
                contig_name[:100]
            )  # Avoid very long filenames
            contig_ordinal_inds[busco_group] += 1
            output_index = contig_ordinal_inds[busco_group]
            out_filename = os.path.join(
                self.pred_genes_dir, "{}.out.{}".format(busco_group, output_index)
            )
            njobs += 1

            yield busco_group, contig_tmp_file, contig_start, contig_end, out_filename

    @log(
        "Additional parameters for Augustus are {}: ",
        logger,
        attr_name="_target_species",
    )
    def parse_parameters(self):
        accepted_keys = []
        accepted_values = []
        if self.extra_params:
            self.extra_params = self.extra_params.strip("\" '")
            try:
                if self.extra_params.startswith("--"):
                    key_val_pairs = self.extra_params.split(" --")
                    for kv in key_val_pairs:
                        key_vals = kv.strip("- ").split("=")
                        if len(key_vals) == 2:
                            key, val = key_vals
                            if key in type(self).ACCEPTED_PARAMETERS:
                                accepted_keys.append(key.strip())
                                accepted_values.append(val.strip())
                            else:
                                logger.warning(
                                    "{} is not an accepted parameter for Augustus.".format(
                                        key
                                    )
                                )
                        else:
                            raise AugustusParsingError
                else:
                    raise AugustusParsingError
            except AugustusParsingError:
                logger.warning(
                    "Augustus parameters are not correctly formatted. Please enter them as follows: "
                    '"--param1=value1 --param2=value2" etc. Proceeding without additional parameters.'
                )
                return [], []
        return accepted_keys, accepted_values

    def _merge_stderr_logs(self):
        with open(self.err_logfile, "a") as f:
            for err_logfile in self.err_logfiles:
                with open(err_logfile, "r") as g:
                    content = g.readlines()
                    f.writelines(content)
        return

    def _remove_individual_err_logs(self):
        shutil.rmtree(self.tmp_dir)
        return

    def get_version(self):  # todo: need to handle all possible exceptions
        augustus_help_output = subprocess.check_output(
            [self.cmd, "--version"], stderr=subprocess.STDOUT, shell=False
        )
        augustus_help_output = augustus_help_output.decode("utf-8")
        s = augustus_help_output.split("\n")[0]
        augustus_version = s[s.find("(") + 1 : s.find(")")]
        return augustus_version

    def configure_job(
        self, busco_group, contig_tmp_file, contig_start, contig_end, out_filename
    ):
        # Augustus does not provide an option to write to an output file, so have to change the pipe target from the
        # log file to the desired output file
        self.logfile_path_out = out_filename
        err_logfile = os.path.join(
            self.tmp_dir, os.path.basename(out_filename.rpartition(".out")[0] + ".err")
        )
        self.logfile_path_err = err_logfile
        self.err_logfiles.append(err_logfile)

        augustus_job = self.create_job()
        augustus_job.add_parameter("--codingseq=1")
        augustus_job.add_parameter(
            "--proteinprofile={}".format(
                os.path.join(
                    self.lineage_dataset, "prfl", "{}.prfl".format(busco_group)
                )
            )
        )
        augustus_job.add_parameter("--predictionStart={}".format(contig_start))
        augustus_job.add_parameter("--predictionEnd={}".format(contig_end))
        augustus_job.add_parameter("--species={}".format(self._target_species))
        for k, key in enumerate(self.param_keys):
            augustus_job.add_parameter("--{}={}".format(key, self.param_values[k]))
        augustus_job.add_parameter(os.path.join(self.seqs_path, contig_tmp_file))
        return augustus_job

    def _extract_genes_from_augustus_output(self, filename):
        # todo: consider parallelizing this and other parsing functions

        gene_id = None
        gene_info = []
        sequences_aa = []
        sequences_nt = []
        gene_found = False
        completed_record = False

        with open(
            os.path.join(self.pred_genes_dir, filename), "r", encoding="utf-8"
        ) as f:
            # utf-8 encoding needed to handle the umlaut in the third line of the file.
            gene_info_section = False
            nt_sequence_section = False
            aa_sequence_section = False
            nt_sequence_parts = []
            aa_sequence_parts = []

            for line in f:

                if aa_sequence_section:
                    if "]" in line:
                        line = line.strip().lstrip("# ").rstrip("]")
                        aa_sequence_parts.append(line)
                        aa_sequence_section = False
                        completed_record = True
                        if gene_id is not None:
                            aa_sequence = "".join(aa_sequence_parts)
                            nt_sequence = "".join(nt_sequence_parts)
                            seq_record_aa = SeqRecord(
                                Seq(aa_sequence.upper()), id=gene_id
                            )
                            seq_record_nt = SeqRecord(
                                Seq(nt_sequence.upper()), id=gene_id
                            )
                            sequences_aa.append(seq_record_aa)
                            sequences_nt.append(seq_record_nt)
                            aa_sequence_parts = []
                            nt_sequence_parts = []
                            gene_id = None
                        continue

                    else:
                        line = line.strip().lstrip("# ").rstrip("]")
                        aa_sequence_parts.append(line)
                        continue

                if line.startswith("# protein"):
                    nt_sequence_section = False
                    aa_sequence_section = True
                    if "]" in line:
                        line = line.strip().rstrip("]").split("[")
                        aa_sequence_parts.append(line[1])
                        aa_sequence_section = False
                        completed_record = True
                        if gene_id is not None:
                            aa_sequence = "".join(aa_sequence_parts)
                            nt_sequence = "".join(nt_sequence_parts)
                            seq_record_aa = SeqRecord(
                                Seq(aa_sequence.upper()), id=gene_id
                            )
                            seq_record_nt = SeqRecord(
                                Seq(nt_sequence.upper()), id=gene_id
                            )
                            sequences_aa.append(seq_record_aa)
                            sequences_nt.append(seq_record_nt)
                            aa_sequence_parts = []
                            nt_sequence_parts = []
                            gene_id = None
                    else:
                        line = line.strip().rstrip("]").split("[")
                        aa_sequence_parts.append(line[1])
                    continue

                if nt_sequence_section:
                    line = line.strip().lstrip("# ").rstrip("]")
                    nt_sequence_parts.append(line)
                    continue

                if line.startswith("# coding sequence"):
                    gene_info = []
                    gene_info_section = False
                    nt_sequence_section = True
                    line = (
                        line.strip().rstrip("]").split("[")
                    )  # Extract sequence part of line
                    nt_sequence_parts.append(line[1])
                    continue

                if gene_info_section:
                    line = line.strip().split()
                    seq_name = line[0].strip()
                    gene_start = line[3].strip()
                    gene_end = line[4].strip()
                    strand = line[6].strip()
                    if not gene_id:
                        gene_id = "{}:{}-{}".format(seq_name, gene_start, gene_end)
                        self.gene_details[gene_id].append(
                            {
                                "gene_start": gene_start,
                                "gene_end": gene_end,
                                "strand": strand,
                            }
                        )
                    gene_info.append("\t".join(line))
                    continue

                if line.startswith("# start gene"):
                    gene_found = True
                    self.any_gene_found = True
                    gene_info_section = True
                    completed_record = False
                    continue

            if gene_found and not completed_record:
                logger.warning("Augustus output file {} truncated".format(filename))

        self.sequences_aa.update({record.id: record for record in sequences_aa})
        self.sequences_nt.update({record.id: record for record in sequences_nt})
        if gene_found:
            self._write_sequences_to_file(filename, sequences_nt, sequences_aa)

        return

    def make_gff_files(self, single_copy_buscos):

        for b in single_copy_buscos:
            gene_info = []
            busco_files = [
                f for f in os.listdir(self.pred_genes_dir) if f.startswith(b)
            ]
            pred_genes_dir_current = self.pred_genes_dir
            if (
                len(busco_files) == 0
                and pred_genes_dir_current == self.pred_genes_dir_rerun
            ):
                pred_genes_dir_current = self.pred_genes_dir_initial
                busco_files = [
                    f for f in os.listdir(pred_genes_dir_current) if f.startswith(b)
                ]
            gff_filename = os.path.join(self.gff_dir, "{}.gff".format(b))
            single_copy_busco_gene = list(single_copy_buscos[b].keys())[0]
            gene_id_parts = single_copy_busco_gene.split(":")
            if (
                len(gene_id_parts) > 2
            ):  # if a ":" is present in the gene id, we don't want to break it up
                gene_id_parts = [":".join(gene_id_parts[:-1]), gene_id_parts[-1]]
            single_copy_busco_gene_id = gene_id_parts[0]
            (
                single_copy_busco_gene_start_coord,
                single_copy_busco_gene_end_coord,
            ) = gene_id_parts[1].split("-")
            gene_found = False
            for filename in busco_files:
                match_number = filename.split(".")[-1]
                with open(
                    os.path.join(pred_genes_dir_current, filename),
                    "r",
                    encoding="utf-8",
                ) as f:
                    gene_info_section = False
                    for line in f:
                        if gene_info_section and line.startswith("# coding sequence"):
                            with open(gff_filename, "w") as g:
                                g.write("\n".join(gene_info) + "\n")
                            gene_info = []
                            break

                        if line.startswith("# start gene"):
                            gene_info_section = True
                            continue

                        if gene_info_section:
                            line = line.strip().split()
                            seq_name = line[0]
                            gene_start = line[3]
                            gene_end = line[4]
                            if gene_found or (
                                seq_name == single_copy_busco_gene_id
                                and gene_start == single_copy_busco_gene_start_coord
                                and gene_end == single_copy_busco_gene_end_coord
                            ):
                                gene_found = True
                                gene_id_info = line[-1]
                                line[-1] = self.edit_gene_identifier(
                                    gene_id_info, match_number
                                )
                                if len(line) == 12:
                                    gene_id_info_2 = line[-3]
                                    line[-3] = self.edit_gene_identifier(
                                        gene_id_info_2, match_number
                                    )
                                gene_info.append("\t".join(line))
                            else:
                                gene_info_section = False
                            continue
                if gene_found:
                    break
            if not gene_found and self.run_number == 1:
                raise SystemExit(
                    "Unable to find single copy BUSCO gene in Augustus output."
                )

        return

    def edit_gene_identifier(self, orig_str, match_num):
        modified_str = re.sub(
            r"g([0-9])", r"r{}.m{}.g\1".format(self.run_number, match_num), orig_str
        )
        return modified_str

    def _write_sequences_to_file(self, filename, sequences_nt, sequences_aa):

        filename_parts = filename.rpartition(".out")
        output_fna = os.path.join(
            self.extracted_prot_dir, filename_parts[0] + ".fna" + filename_parts[-1]
        )
        output_faa = os.path.join(
            self.extracted_prot_dir, filename_parts[0] + ".faa" + filename_parts[-1]
        )
        self.output_sequences.append(output_faa)

        with open(output_fna, "w") as out_fna:
            SeqIO.write(sequences_nt, out_fna, "fasta")
        with open(output_faa, "w") as out_faa:
            SeqIO.write(sequences_aa, out_faa, "fasta")

        return

    def move_retraining_parameters(self):
        """
        This function moves retraining parameters from augustus species folder
        to the run folder
        """
        augustus_species_path = os.path.join(
            self._augustus_config_path, "species", self._target_species
        )
        if os.path.exists(augustus_species_path):
            new_path = os.path.join(
                self._output_folder, "retraining_parameters", self._target_species
            )
            shutil.move(augustus_species_path, new_path)
        elif self.config.getboolean("busco_run", "restart") and os.path.exists(
            os.path.join(
                self._output_folder, "retraining_parameters", self._target_species
            )
        ):
            pass
        else:
            logger.warning("Augustus did not produce a retrained species folder.")
        return


class GFF2GBRunner(BaseRunner):

    name = "gff2gbSmallDNA.pl"
    cmd = "gff2gbSmallDNA.pl"

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(self.run_folder, "augustus_output")
        self.gff_folder = os.path.join(self._output_folder, "gff")
        self.gb_folder = os.path.join(self._output_folder, "gb")
        self.create_dirs([self.gff_folder, self.gb_folder])

        self.init_checkpoint_file()

    def configure_runner(self, single_copy_buscos):
        self.run_number += 1
        self.single_copy_buscos = single_copy_buscos

    def run(self):
        super().run()
        self.total = self._count_jobs()
        self.run_jobs()

    def _count_jobs(self):
        n = len(self.single_copy_buscos)
        return n

    def generate_job_args(self):
        for busco_id in self.single_copy_buscos:
            yield busco_id

    def configure_job(self, busco_id):
        gff2_gb_small_dna_pl_job = self.create_job()
        gff2_gb_small_dna_pl_job.add_parameter(
            os.path.join(self.gff_folder, "{}.gff".format(busco_id))
        )
        gff2_gb_small_dna_pl_job.add_parameter(self.input_file)
        gff2_gb_small_dna_pl_job.add_parameter("1000")
        gff2_gb_small_dna_pl_job.add_parameter(
            os.path.join(self.gb_folder, "{}.raw.gb".format(busco_id))
        )
        return gff2_gb_small_dna_pl_job

    def check_tool_dependencies(self):
        pass

    def get_version(self):
        return

    @property
    def output_folder(self):
        return self._output_folder


class NewSpeciesRunner(BaseRunner):

    name = "new_species.pl"
    cmd = "new_species.pl"

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(self.run_folder, "augustus_output")
        self.new_species_name = "BUSCO_{}".format(os.path.basename(self.main_out))
        self.init_checkpoint_file()
        self.run_number += 1

    def run(self):
        super().run()
        self.total = 1
        self.run_jobs()

    def configure_job(self, *args):

        new_species_pl_job = self.create_job()
        # bacteria clade needs to be flagged as "prokaryotic"
        if self.domain == "prokaryota":
            new_species_pl_job.add_parameter("--prokaryotic")
        new_species_pl_job.add_parameter(
            "--species={}".format(os.path.basename(self.new_species_name))
        )
        return new_species_pl_job

    def check_tool_dependencies(self):
        pass

    def generate_job_args(self):
        yield

    def get_version(self):
        return

    @property
    def output_folder(self):
        return self._output_folder


class ETrainingRunner(BaseRunner):

    name = "etraining"
    cmd = "etraining"

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(self.run_folder, "augustus_output")
        self._gb_folder = os.path.join(self._output_folder, "gb")
        self.augustus_config_path = self.config.get("busco_run", "augustus_config_path")
        self._training_file = os.path.join(self._output_folder, "training_set.db")

        self.init_checkpoint_file()

    def configure_runner(self, new_species_name):
        self.run_number += 1
        self.new_species_name = new_species_name
        self._merge_gb_files()

    def run(self):
        super().run()
        self.total = 1
        self.run_jobs()
        self._validate_run()

    def check_tool_dependencies(self):
        pass

    def generate_job_args(self):
        yield

    def _merge_gb_files(self):
        """Concatenate all GB files into one large file"""
        with open(self._training_file, "w") as outfile:
            for fname in os.listdir(self._gb_folder):
                with open(os.path.join(self._gb_folder, fname), "r") as infile:
                    outfile.writelines(infile.readlines())
        return

    def _validate_run(self):
        species_filepath = os.path.join(
            self.augustus_config_path, "species", self.new_species_name
        )
        if os.path.exists(species_filepath) and any(
            "exon_probs" in f for f in os.listdir(species_filepath)
        ):
            return
        else:
            raise SystemExit(
                "Retraining did not complete correctly. Check your Augustus config path environment variable."
            )

    def configure_job(self, *args):
        etraining_job = self.create_job()
        etraining_job.add_parameter("--species={}".format(self.new_species_name))
        etraining_job.add_parameter(
            os.path.join(self.run_folder, "augustus_output", "training_set.db")
        )
        return etraining_job

    def get_version(self):
        return

    @property
    def output_folder(self):
        return self._output_folder


class OptimizeAugustusRunner(BaseRunner):

    name = "optimize_augustus.pl"
    cmd = "optimize_augustus.pl"

    def __init__(self):
        super().__init__()
        self._output_folder = None
        self.training_set_db = None
        self.new_species_name = None

    def configure_runner(self, output_folder, new_species_name):
        self.run_number += 1
        self._output_folder = output_folder
        self.training_set_db = os.path.join(self._output_folder, "training_set.db")
        self.new_species_name = new_species_name

        self.init_checkpoint_file()

    def configure_job(self, *args):
        optimize_augustus_pl_job = self.create_job()
        optimize_augustus_pl_job.add_parameter("--cpus={}".format(self.cpus))
        optimize_augustus_pl_job.add_parameter(
            "--species={}".format(self.new_species_name)
        )
        optimize_augustus_pl_job.add_parameter(self.training_set_db)
        return optimize_augustus_pl_job

    def run(self):
        super().run()
        self.total = 1
        self.run_jobs()

    def generate_job_args(self):
        yield

    def check_tool_dependencies(self):
        pass

    def get_version(self):
        return

    @property
    def output_folder(self):
        return self._output_folder



from configparser import ConfigParser
from configparser import NoOptionError, NoSectionError
from configparser import ParsingError
from configparser import DuplicateOptionError
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from abc import ABCMeta, abstractmethod
from busco.BuscoDownloadManager import BuscoDownloadManager
import os
import shutil
import glob
import pprint

logger = BuscoLogger.get_logger(__name__)


class BaseConfig(ConfigParser):

    DEFAULT_ARGS_VALUES = {
        "out_path": os.getcwd(),
        "cpu": 1,
        "force": False,
        "restart": False,
        "quiet": False,
        "download_path": os.path.join(os.getcwd(), "busco_downloads"),
        "datasets_version": "odb10",
        "offline": False,
        "download_base_url": "https://busco-data.ezlab.org/v5/data/",
        "auto-lineage": False,
        "auto-lineage-prok": False,
        "auto-lineage-euk": False,
        "update-data": False,
        "evalue": 1e-3,
        "limit": 3,
        "use_augustus": False,
        "long": False,
    }

    DEPENDENCY_SECTIONS = {
        "tblastn",
        "makeblastdb",
        "prodigal",
        "sepp",
        "metaeuk",
        "augustus",
        "etraining",
        "gff2gbSmallDNA.pl",
        "new_species.pl",
        "optimize_augustus.pl",
        "hmmsearch",
    }

    def __init__(self):
        super().__init__()
        config_dict = {"busco_run": type(self).DEFAULT_ARGS_VALUES}
        config_dict.update(
            {
                tool: {"path": "", "command": ""}
                for tool in type(self).DEPENDENCY_SECTIONS
            }
        )
        self.read_dict(config_dict)

    def _load_config_file(self):
        """
        Load config file using ConfigParser.
        :return:
        """
        try:
            with open(self.conf_file) as cfg_file:
                self.read_file(cfg_file)
        except IOError:
            raise SystemExit("Config file {} cannot be found".format(self.conf_file))
        except ParsingError:
            raise SystemExit(
                "Unable to parse the contents of config file {}".format(self.conf_file)
            )
        except DuplicateOptionError:
            raise SystemExit(
                "Duplicated entry in config file {}. Unable to load configuration.".format(
                    self.conf_file
                )
            )
        return

    def _init_downloader(self):
        """
        Initialize BuscoDownloadManager to control any file downloads from the BUSCO server.
        :return:
        """
        self.downloader = BuscoDownloadManager(self)
        return

    # @log("Setting value in config")
    # def set(self, *args, **kwargs):
    #     super().set(*args, **kwargs)


class PseudoConfig(BaseConfig):
    def __init__(self, conf_file):
        super().__init__()
        self.conf_file = conf_file

    def load(self):
        if self.conf_file != "local environment":
            self._load_config_file()
        else:
            self.set("busco_run", "update-data", "True")
        self._fill_default_values()
        self._init_downloader()

    def _fill_default_values(self):
        self.set("busco_run", "offline", "False")

        try:
            self.get("busco_run", "download_base_url")
        except NoOptionError:
            self.set(
                "busco_run",
                "download_base_url",
                type(self).DEFAULT_ARGS_VALUES["download_base_url"],
            )

        try:
            self.get("busco_run", "download_path")
        except NoOptionError:
            self.set(
                "busco_run",
                "download_path",
                type(self).DEFAULT_ARGS_VALUES["download_path"],
            )

        try:
            self.update = self.getboolean("busco_run", "update-data")
            if not self.update:
                self.existing_downloads = sorted(
                    glob.glob(
                        os.path.join(
                            self.get("busco_run", "download_path"),
                            "information",
                            "lineages_list*.txt",
                        )
                    )
                )[::-1]
                if self.existing_downloads:
                    logger.warning(
                        "The datasets list shown may not be up-to-date. To get current information, make sure "
                        "you have set 'update-data=True' in your config file."
                    )
                else:
                    raise SystemExit(
                        "Unable to download list of datasets. Please make sure you have set "
                        "update-data=True in your config file."
                    )

        except NoOptionError:
            self.set("busco_run", "update-data", "True")


class BuscoConfig(ConfigParser, metaclass=ABCMeta):
    """
    This class extends busco.PipeConfig to read the config.ini file. Furthermore, it uses extra args that can be
    provided through command line and information available in the dataset.cfg file to produce a single instance
    containing all correct parameters to be injected to a busco.BuscoAnalysis instance.
    """

    FORBIDDEN_HEADER_CHARS = [
        "ç",
        "¬",
        "¢",
        "´",
        "ê",
        "î",
        "ô",
        "ŵ",
        "ẑ",
        "û",
        "â",
        "ŝ",
        "ĝ",
        "ĥ",
        "ĵ",
        "ŷ",
        "ĉ",
        "é",
        "ï",
        "ẅ",
        "ë",
        "ẅ",
        "ë",
        "ẗ,",
        "ü",
        "í",
        "ö",
        "ḧ",
        "é",
        "ÿ",
        "ẍ",
        "è",
        "é",
        "à",
        "ä",
        "¨",
        "€",
        "£",
        "á",
    ]

    FORBIDDEN_HEADER_CHARS_BEFORE_SPLIT = ["/", '"']

    HMMER_VERSION = 3.1

    def __init__(self, **kwargs):

        super().__init__(**kwargs)

    def download_lineage_file(self, lineage):
        """
        Download lineage dataset if not present using BuscoDownloadManager.
        :param str lineage: Basename of the lineage dataset
        :return str lineage_filepath: Local path to downloaded file
        """
        local_lineage_filepath = self.downloader.get(lineage, "lineages")
        self.set("busco_run", "lineage_dataset", local_lineage_filepath)
        return

    def load_dataset_config(self):
        """
        Load BUSCO dataset config file.
        :return:
        """
        try:
            with open(
                os.path.join(self.get("busco_run", "lineage_dataset"), "dataset.cfg"),
                "r",
            ) as target_species_file:
                dataset_kwargs = dict(
                    line.strip().split("=") for line in target_species_file
                )
                for key, value in dataset_kwargs.items():
                    if key == "species":
                        try:
                            config_species = self.get("busco_run", "augustus_species")
                            if config_species != value:
                                logger.warning(
                                    "An augustus species was mentioned in the config file or on the command "
                                    "line, dataset default species ({}) will be ignored".format(
                                        value
                                    )
                                )
                        except NoOptionError:
                            self.set("busco_run", "augustus_species", value)

                    if key in [
                        "prodigal_genetic_code",
                        "ambiguous_cd_range_upper",
                        "ambiguous_cd_range_lower",
                    ]:
                        self.set("prodigal", key, value)
                    else:
                        self.set("busco_run", key, value)

        except IOError:
            logger.warning(
                "The dataset you provided does not contain the file dataset.cfg and is not valid for "
                "BUSCO v4.0 or higher"
            )
        return

    @abstractmethod
    def _create_required_paths(self, main_out):
        """
        Mandatory child class method. Overridden by child classes.
        :return:
        """
        if not os.path.exists(main_out):
            os.makedirs(main_out)

    def set_results_dirname(self, lineage):
        self.set(
            "busco_run",
            "lineage_results_dir",
            "run_{}".format(os.path.basename(lineage)),
        )
        return


class BuscoConfigAuto(BuscoConfig):
    def __init__(self, config, lineage, **kwargs):
        super().__init__(**kwargs)
        self._propagate_config(config)
        self.set_results_dirname(lineage)

        self._create_required_paths()
        self.download_lineage_file(lineage)
        self.load_dataset_config()

    def _create_required_paths(self):
        """
        Create directory for auto-lineage runs.
        :return:
        """
        main_out = os.path.join(self.get("busco_run", "main_out"), "auto_lineage")
        super()._create_required_paths(main_out)
        return

    def _propagate_config(self, config):
        """
        Copy all information from BuscoConfigMain sections into this BuscoConfigAuto object.
        Also copy BuscoDownloadManager object instead of instantiating a second one.
        :param config:
        :return:
        """
        for key, value in config.items():
            self[key] = value

        self.downloader = config.downloader
        return


class BuscoConfigMain(BuscoConfig, BaseConfig):

    MANDATORY_USER_PROVIDED_PARAMS = ["in", "out", "mode"]

    PERMITTED_OPTIONS = [
        "in",
        "out",
        "out_path",
        "mode",
        "auto-lineage",
        "auto-lineage-prok",
        "auto-lineage-euk",
        "cpu",
        "force",
        "restart",
        "download_path",
        "datasets_version",
        "quiet",
        "offline",
        "long",
        "augustus_parameters",
        "augustus_species",
        "download_base_url",
        "lineage_dataset",
        "update-data",
        "metaeuk_parameters",
        "metaeuk_rerun_parameters",
        "evalue",
        "limit",
        "use_augustus",
    ]

    def __init__(self, conf_file, params, **kwargs):
        """
        :param conf_file: a path to a config.ini file
        :type conf_file: str
        :param params: key and values matching BUSCO parameters to override config.ini values
        :type params: dict
        """
        super().__init__(**kwargs)
        self.conf_file = conf_file
        self.params = params
        self.main_out = None

    def configure(self):
        if self.conf_file != "local environment":
            self._load_config_file()
        # Update the config with args provided by the user, else keep config
        self._update_config_with_args(self.params)
        self._check_value_constraints()

    @classmethod
    def merge_two_dicts(cls, x, y):
        """Given two dictionaries, merge them into a new dict as a shallow copy."""
        z = x.copy()
        z.update(y)
        return z

    def validate(self):
        self._check_mandatory_keys_exist()
        self._check_no_previous_run()
        self._check_allowed_keys()
        self._create_required_paths()
        self._cleanup_config()
        self._check_required_input_exists()

        self._init_downloader()

        self.log_config()

    def log_config(self):
        logger.debug("State of BUSCO config before run:")
        logger.debug(PrettyLog(vars(self)))

    def check_lineage_present(self):
        """
        Check if "lineage_dataset" parameter has been provided by user.
        :return: True if present, False if not
        :rtype: bool
        """
        try:
            lineage_dataset = self.get("busco_run", "lineage_dataset")
            datasets_version = self.get("busco_run", "datasets_version")
            if "_odb" in lineage_dataset:
                dataset_version = lineage_dataset.rsplit("_")[-1].rstrip("/")
                if datasets_version != dataset_version:
                    logger.warning(
                        "There is a conflict in your config. You specified a dataset from {0} while "
                        "simultaneously requesting the datasets_version parameter be {1}. Proceeding with "
                        "the lineage dataset as specified from {0}".format(
                            dataset_version, datasets_version
                        )
                    )
                self.set("busco_run", "datasets_version", dataset_version)
            else:  # Make sure the ODB version is in the dataset name
                lineage_dataset = "_".join([lineage_dataset, datasets_version])
                self.set("busco_run", "lineage_dataset", lineage_dataset)

            datasets_version = self.get("busco_run", "datasets_version")
            if datasets_version != "odb10":
                raise SystemExit(
                    "BUSCO v4 only works with datasets from OrthoDB v10 (with the suffix '_odb10'). "
                    "For a full list of available datasets, enter 'busco --list-datasets'. "
                    "You can also run BUSCO using auto-lineage, to allow BUSCO to automatically select "
                    "the best dataset for your input data."
                )
            return True
        except NoOptionError:  # todo: need a helpful error message if datasets_version is not set but lineage_dataset is.
            return False

    def _check_evalue(self):
        """
        Warn the user if the config contains a non-standard e-value cutoff.
        :return:
        """
        if (
            self.getfloat("busco_run", "evalue")
            != type(self).DEFAULT_ARGS_VALUES["evalue"]
        ):  # todo: introduce systemexit if not float
            logger.warning("You are using a custom e-value cutoff")
        return

    def _check_limit_value(self):
        """
        Check the value of limit. Ensure it is between 1 and 20, otherwise raise SystemExit.
        :return:
        """
        limit_val = self.getint(
            "busco_run", "limit"
        )  # todo: introduce systemexit if not int.
        if limit_val <= 0 or limit_val > 20:
            raise SystemExit(
                "Limit must be an integer between 1 and 20 (you have used: {}). Note that this parameter "
                "is not needed by the protein mode.".format(limit_val)
            )
        return

    @log("Mode is {0}", logger, attr_name="_mode", on_func_exit=True, log_once=True)
    def _check_mandatory_keys_exist(self):
        """
        Make sure all mandatory user-provided parameters are present in the config.
        :return:
        """
        for param in type(self).MANDATORY_USER_PROVIDED_PARAMS:
            try:
                value = self.get("busco_run", param)
                if param == "mode":
                    synonyms = {
                        "genome": [
                            "genome",
                            "geno",
                            "genomes",
                            "Genome",
                            "Genomes",
                            "Geno",
                        ],
                        "transcriptome": [
                            "transcriptome",
                            "tran",
                            "transcriptomes",
                            "trans",
                            "Transcriptome",
                            "Transcriptomes",
                            "Tran",
                            "Trans",
                        ],
                        "proteins": [
                            "proteins",
                            "prot",
                            "protein",
                            "Proteins",
                            "Protein",
                            "Prot",
                            "proteome",
                            "proteomes",
                            "Proteome",
                            "Proteomes",
                        ],
                    }
                    if value not in list(
                        synonyms["genome"]
                        + synonyms["transcriptome"]
                        + synonyms["proteins"]
                    ):
                        raise SystemExit(
                            "Unknown mode {}.\n'Mode' parameter must be one of "
                            "['genome', 'transcriptome', 'proteins']".format(value)
                        )
                    if value in synonyms["genome"]:
                        self.set("busco_run", "mode", "genome")
                    elif value in synonyms["transcriptome"]:
                        self.set("busco_run", "mode", "transcriptome")
                    elif value in synonyms["proteins"]:
                        self.set("busco_run", "mode", "proteins")

                    self._mode = self.get("busco_run", "mode")

                if param == "out":
                    if value == "":
                        raise SystemExit(
                            "Please specify an output name for the BUSCO run. "
                            "This can be done using the -o flag or in the config file"
                        )
            except NoOptionError:
                raise SystemExit(
                    'The parameter "{} (--{})" was not provided. '
                    "Please add it in the config file or provide it "
                    "through the command line".format(param, param)
                )
        return

    def _check_allowed_keys(self):
        full_dict = {"busco_run": type(self).PERMITTED_OPTIONS}
        full_dict.update(
            {
                dependency: ["path", "command"]
                for dependency in type(self).DEPENDENCY_SECTIONS
            }
        )

        for section_name, options in full_dict.items():
            try:
                for option in self.options(section_name):
                    if option not in options:
                        raise SystemExit(
                            "Unrecognized option '{}' in config file".format(option)
                        )
            except NoSectionError:
                logger.warning("Section {} not found".format(section_name))
        return

    def _check_out_value(self):
        """
        Prevent the user form using "/" in out name
        :return:
        """
        if "/" in self.get("busco_run", "out"):
            raise SystemExit(
                "Please do not provide a full path in --out parameter, no slash. "
                "Use out_path in the config.ini file to specify the full path."
            )
        return

    @log(
        "Input file is {}",
        logger,
        attr_name="_input_filepath",
        on_func_exit=True,
        log_once=True,
    )
    def _check_required_input_exists(self):
        """
        Test for existence of input file.
        :return:
        """
        self._input_filepath = self.get("busco_run", "in")
        if not os.path.exists(self._input_filepath):
            raise SystemExit(
                "Input file {} does not exist".format(self._input_filepath)
            )
        return

    def _check_no_previous_run(self):
        self.main_out = os.path.join(
            self.get("busco_run", "out_path"), self.get("busco_run", "out")
        )
        if os.path.exists(self.main_out):
            if self.getboolean("busco_run", "force"):
                self._force_remove_existing_output_dir(self.main_out)
            elif self.getboolean("busco_run", "restart"):
                logger.info(
                    "Attempting to restart the run using the following directory: {}".format(
                        self.main_out
                    )
                )
            else:
                raise SystemExit(
                    "A run with the name {} already exists...\n"
                    "\tIf you are sure you wish to overwrite existing files, "
                    "please use the -f (force) option".format(self.main_out)
                )
        elif self.getboolean("busco_run", "restart"):
            logger.warning(
                "Restart mode not available as directory {} does not exist.".format(
                    self.main_out
                )
            )
            self.set("busco_run", "restart", "False")

        return

    def _cleanup_config(self):
        """
        Collection of housekeeping functions to ensure configuration is suitable.
        :return:
        """
        self._check_out_value()
        self._check_limit_value()
        self._check_evalue()
        self._expand_all_paths()

    @staticmethod
    @log(
        "'Force' option selected; overwriting previous results directory", logger
    )  # todo: review log messages
    def _force_remove_existing_output_dir(dirpath):
        """
        Remove main output folder from a previous BUSCO run.
        :return:
        """
        shutil.rmtree(dirpath)
        return

    def _create_required_paths(self):
        """
        Create main output directory and tmp directory.
        :return:
        """
        super()._create_required_paths(self.main_out)
        self.set("busco_run", "main_out", self.main_out)
        return

    def _expand_all_paths(self):
        """
        Convert relative pathnames beginning with "~" or "." into absolute paths.
        :return:
        """
        for key in self.sections():
            for item in self.items(key):
                if item[0].endswith("_path") or item[0] == "path" or item[0] == "in":
                    if item[1].startswith("~"):
                        self.set(key, item[0], os.path.expanduser(item[1]))
                    elif item[1].startswith(".") or (item[1] and "/" not in item[1]):
                        self.set(key, item[0], os.path.abspath(item[1]))

        return

    def _check_value_constraints(self):
        """
        Load default values into config if not provided in config file or on the command line.
        :return:
        """
        # for param in list(type(self).DEFAULT_ARGS_VALUES.keys()):
        #     try:
        #         self.get("busco_run", param)
        #     except NoOptionError:
        #         self.set("busco_run", param, str(type(self).DEFAULT_ARGS_VALUES[param]))

        # Set auto-lineage to True if either auto-lineage-prok or auto-lineage-euk is selected
        if self.getboolean("busco_run", "auto-lineage-prok") or self.getboolean(
            "busco_run", "auto-lineage-euk"
        ):
            self.set("busco_run", "auto-lineage", "True")

        if self.getboolean("busco_run", "auto-lineage-prok") and self.getboolean(
            "busco_run", "auto-lineage-euk"
        ):
            logger.warning(
                "You have specified both --auto-lineage-prok and --auto-lineage-euk. This has the same behaviour as --auto-lineage."
            )
            self.set("busco_run", "auto-lineage-prok", "False")
            self.set("busco_run", "auto-lineage-euk", "False")

        return

    def _update_config_with_args(self, args):
        """
        Include command line arguments in config. Overwrite any values given in the config file.
        :param args: Dictionary of parsed command line arguments. To see full list, inspect run_BUSCO_unittests.py or
        type busco -h
        :type args: dict
        :return:
        """
        for key, val in args.items():
            if key in type(self).PERMITTED_OPTIONS:
                if val is not None and type(val) is not bool:
                    self.set("busco_run", key, str(val))
                elif val:  # if True
                    self.set("busco_run", key, "True")
        return


# Code taken from https://dave.dkjones.org/posts/2013/pretty-print-log-python/
class PrettyLog:
    def __init__(self, obj):
        self.obj = obj

    def __repr__(self):
        return pprint.pformat(self.obj)



import os
from busco.BuscoLogger import BuscoLogger
from busco.busco_tools.Toolset import Tool
from shutil import which
from abc import ABCMeta, abstractmethod
from busco.BuscoConfig import BuscoConfigAuto
import time

logger = BuscoLogger.get_logger(__name__)


class ToolException(Exception):
    """
    Module-specific exception
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


class BaseRunner(Tool, metaclass=ABCMeta):

    config = None

    def __init__(self):
        super().__init__()
        self.run_number = 0
        self.input_file = self.config.get("busco_run", "in")
        self.main_out = self.config.get("busco_run", "main_out")
        self.working_dir = (
            os.path.join(self.main_out, "auto_lineage")
            if isinstance(self.config, BuscoConfigAuto)
            else self.main_out
        )
        self.lineage_results_dir = self.config.get("busco_run", "lineage_results_dir")
        self.run_folder = os.path.join(self.working_dir, self.lineage_results_dir)
        self.log_folder = os.path.join(self.main_out, "logs")
        self.cpus = self.config.getint("busco_run", "cpu")
        self.lineage_dataset = self.config.get("busco_run", "lineage_dataset")
        self.domain = self.config.get("busco_run", "domain")

        if not self.check_tool_available():
            raise ToolException(
                "{} tool cannot be found. Please check the 'path' and 'command' parameters "
                "provided in the config file. Do not include the command in the "
                "path!".format(self.name)
            )
        self.version = self.get_version()
        self.check_tool_dependencies()

        self.checkpoint_file = None

        self.logfile_path_out = os.path.join(
            self.config.get("busco_run", "main_out"),
            "logs",
            "{}_out.log".format(self.name),
        )
        self.logfile_path_err = (
            self.logfile_path_out.rpartition("_out.log")[0] + "_err.log"
        )

    def init_checkpoint_file(self):
        self.checkpoint_file = os.path.join(self.output_folder, ".checkpoint")

    def write_checkpoint_file(self):
        with open(self.checkpoint_file, "a") as cpt_file:
            cpt_file.write("Tool: {}\n".format(self.name))
            cpt_file.write("Version: {}\n".format(self.version))
            cpt_file.write("Run: {}\n".format(self.run_number))
            cpt_file.write("Time: {}\n".format(time.strftime("%m/%d/%Y %H:%M:%S")))
            cpt_file.write("Completed {} jobs\n\n".format(self.total))

    def check_previous_completed_run(self):
        if not os.path.exists(self.checkpoint_file):
            return False
        else:
            with open(self.checkpoint_file, "r") as cpt_file:
                lines = cpt_file.readlines()
                tool_names = [s.strip().split(": ")[1] for s in lines[0::6]]
                tool_versions = [s.strip().split(": ")[1] for s in lines[1::6]]
                tool_run_numbers = [s.strip().split(": ")[1] for s in lines[2::6]]
                try:
                    start_search = 0
                    while True:
                        tool_ind = tool_names.index(self.name, start_search)
                        if str(self.version) != str(tool_versions[tool_ind]):
                            logger.warning(
                                "A previous run used {} version {}. "
                                "The restarted run is using {} version "
                                "{}".format(
                                    self.name,
                                    tool_versions[tool_ind],
                                    self.name,
                                    self.version,
                                )
                            )
                        if int(tool_run_numbers[tool_ind]) == int(self.run_number):
                            return True
                        elif int(tool_run_numbers[tool_ind]) < int(self.run_number):
                            start_search = tool_ind + 1
                        else:
                            raise SystemExit(
                                "Something went wrong. Information for {} run {} missing but "
                                "information for run {} found.".format(
                                    self.name,
                                    self.run_number,
                                    tool_run_numbers[tool_ind],
                                )
                            )

                except ValueError:
                    return False

                except TypeError:
                    logger.warning(
                        "Unable to parse {} file. Restart mode not available.".format(
                            self.checkpoint_file
                        )
                    )

    @abstractmethod
    def check_tool_dependencies(self):
        pass

    @abstractmethod
    def configure_job(self, *args):
        pass

    @abstractmethod
    def generate_job_args(self):
        pass

    @property
    @abstractmethod
    def output_folder(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError

    @abstractmethod
    def run(self):
        if self.version is not None:
            logger.debug("Tool: {}".format(self.name))
            logger.debug("Version: {}".format(self.version))

    @staticmethod
    def create_dirs(dirnames):
        """
        Create all required directories

        :param dirnames: list of paths already constructed
        :return:
        """
        if isinstance(dirnames, str):
            os.makedirs(dirnames, exist_ok=True)
        elif isinstance(dirnames, list):
            for d in dirnames:
                os.makedirs(d, exist_ok=True)
        else:
            raise TypeError("'dirnames' should be either a str or a list")

    def check_tool_available(self):
        """
        Check tool's availability.


        :return: True if the tool can be run, False if it is not the case
        :rtype: bool
        """
        try:
            self.get_tool_from_config()
        except ToolException:
            self.get_tool_from_environment()

        return which(self.cmd) is not None  # True if tool available

    def get_tool_from_environment(self):
        which_tool = which(self.cmd)
        if not which_tool:
            raise ToolException("Tool {} not found".format(self.name))

    def get_tool_from_config(self):
        """
        1. The section ['name'] is available in the config
        2. This section contains keys 'path' and 'command'
        3. The string resulted from concatenation of values of these two keys
        represents the full path to the command
        :return:
        """
        if not self.config.has_section(self.name):
            raise ToolException(
                "Section for the tool [{}] is not present in the config file".format(
                    self.name
                )
            )

        if not self.config.has_option(self.name, "path") or not self.config.get(
            self.name, "path"
        ):
            raise ToolException(
                "Key 'path' in the section [{}] is not present in the config file".format(
                    self.name
                )
            )

        if self.config.has_option(self.name, "command"):
            executable = self.config.get(self.name, "command")
        else:
            executable = self.name

        self.cmd = os.path.join(self.config.get(self.name, "path"), executable)

        return

    @abstractmethod
    def get_version(self):
        return


class NoGenesError(Exception):
    def __init__(self, gene_predictor):
        self.gene_predictor = gene_predictor


class NoRerunFile(Exception):
    def __init__(self):
        pass



#!/usr/bin/env python3
# coding: utf-8
"""
.. module:: Toolset
   :synopsis: the interface to OS enables to run executables / scripts
   in external processes
.. versionadded:: 3.0.0
.. versionchanged:: 4.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""
import os
import subprocess
from subprocess import TimeoutExpired
from multiprocessing import Process, Pool, Value, Lock
import time
from abc import ABCMeta, abstractmethod
from busco.BuscoLogger import BuscoLogger, ToolLogger
from busco.BuscoLogger import LogDecorator as log
from busco.BuscoLogger import StreamLogger
import logging

logger = BuscoLogger.get_logger(__name__)


class Job(Process):
    """
    Build and executes one work item in an external process
    """

    def __init__(
        self, tool_name, cmd, job_outlogger, job_errlogger, timeout, cwd, **kwargs
    ):
        """
        :param name: a name of an executable / script ("a tool") to be run
        :type cmd: list
        :param thread_id: an int id for the thread
        :type thread_id: int
        """
        # initialize parent
        super().__init__()

        self.tool_name = tool_name
        self.cmd_line = [cmd]
        self.job_outlogger = job_outlogger
        self.job_errlogger = job_errlogger
        self.timeout = timeout
        self.cwd = cwd
        self.kwargs = kwargs

    def add_parameter(self, parameter):
        """
        Append parameter to the command line
        :parameter: a parameter
        :type parameter: str
        """
        self.cmd_line.append(parameter)

    @log("cmd call: {}", logger, attr_name="cmd_line", apply="join", debug=True)
    def run(self):
        """
        Start external process and block the current thread's execution
        till the process' run is over
        """
        with StreamLogger(logging.DEBUG, self.job_outlogger, **self.kwargs) as out:
            with StreamLogger(logging.ERROR, self.job_errlogger) as err:
                try:
                    # Stick with Popen(), communicate() and wait() instead of just run() to ensure compatibility with
                    # Python versions < 3.5.
                    p = subprocess.Popen(
                        self.cmd_line, shell=False, stdout=out, stderr=err, cwd=self.cwd
                    )
                    p.wait(self.timeout)
                except TimeoutExpired:
                    p.kill()
                    logger.warning(
                        "The following job was killed as it was taking too long (>1hr) to "
                        "complete.\n{}".format(" ".join(self.cmd_line))
                    )

        self.job_outlogger._file_hdlr.close()
        self.job_outlogger.removeHandler(self.job_outlogger._file_hdlr)
        self.job_errlogger._file_hdlr.close()
        self.job_errlogger.removeHandler(self.job_errlogger._file_hdlr)
        with cnt.get_lock():
            cnt.value += 1


class ToolException(Exception):
    """
    Module-specific exception
    """

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


class Tool(metaclass=ABCMeta):
    """
    Collection of utility methods used by all tools
    """

    def __init__(self):
        """
        Initialize job list for a tool
        :param name: the name of the tool to execute
        :type name: str
        :param config: initialized instance of ConfigParser
        :type config: configparser.ConfigParser
        """
        if self.name == "augustus":
            self.kwargs = {"augustus_out": True}
            self.timeout = 3600
        else:
            self.kwargs = {}
            self.timeout = None
        self.jobs_to_run = []
        self.jobs_running = []
        self.nb_done = 0
        self.total = 0
        self.cpus = None
        self.chunksize = None
        self.cwd = os.getcwd()

    @abstractmethod
    def configure_job(self):
        pass

    @abstractmethod
    def generate_job_args(self):
        pass

    @property
    @abstractmethod
    def name(self):
        raise NotImplementedError

    @abstractmethod
    def write_checkpoint_file(self):
        pass

    def create_job(self):
        """
        Create one work item
        """
        self.tool_outlogger = ToolLogger(self.logfile_path_out)
        self.tool_errlogger = ToolLogger(self.logfile_path_err)
        job = Job(
            self.name,
            self.cmd[:],
            self.tool_outlogger,
            self.tool_errlogger,
            self.timeout,
            self.cwd,
            **self.kwargs
        )
        self.jobs_to_run.append(job)
        # if self.count_jobs_created:
        #     self.total += 1
        return job

    def remove_job(self, job):
        """
        Remove one work item
        :param job: the Job to remove
        :type job: Job
        """
        self.jobs_to_run.remove(job)

    def log_jobs_to_run(self):
        logger.info(
            "Running {} job(s) on {}, starting at {}".format(
                self.total, self.name, time.strftime("%m/%d/%Y %H:%M:%S")
            )
        )
        return

    @log("No jobs to run on {}", logger, attr_name="name", iswarn=True)
    def log_no_jobs(self):
        return

    def run_jobs(self):
        if self.total > 0:
            self.log_jobs_to_run()
        else:
            self.log_no_jobs()
            return

        if (
            self.cpus is None
        ):  # todo: need a different way to ensure self.cpus is nonzero number.
            raise SystemExit("Number of CPUs not specified.")

        with Pool(
            self.cpus, initializer=type(self).init_globals, initargs=(Value("i", 0),)
        ) as job_pool:
            job_pool.map(
                self.run_job, self.generate_job_args(), chunksize=self.chunksize
            )
        self.write_checkpoint_file()

    def run_job(self, args):
        args = (
            (args,) if isinstance(args, str) else tuple(args or (args,))
        )  # Ensure args are tuples that can be unpacked. If no args, args=None, which is falsy, and this evaluates to (None,)
        job = self.configure_job(*args)
        job.run()
        self.nb_done = cnt.value
        if (
            self.nb_done == self.total
            or int(self.nb_done % float(self.total / 10)) == 0
        ):
            self._track_progress()

    @log(
        "[{0}]\t{1} of {2} task(s) completed",
        logger,
        attr_name=["name", "nb_done", "total"],
        on_func_exit=True,
    )
    def _track_progress(self):
        return

    @classmethod
    def init_globals(cls, counter):
        """Counter code adapted from the answer here: https://stackoverflow.com/a/53621343/4844311"""
        global cnt
        cnt = counter



from busco.busco_tools.base import BaseRunner, NoGenesError
import os
import re
from collections import defaultdict
from busco.BuscoLogger import BuscoLogger
from busco.BuscoLogger import LogDecorator as log
from Bio import SeqIO
import shutil
import numpy as np
from configparser import NoOptionError
import subprocess

logger = BuscoLogger.get_logger(__name__)


class ProdigalRunner(BaseRunner):

    name = "prodigal"
    cmd = "prodigal"

    _gc_run_results = defaultdict(dict)

    def __init__(self):
        super().__init__()
        self._output_folder = os.path.join(self.main_out, "prodigal_output")
        self._pred_genes_dir = os.path.join(self._output_folder, "predicted_genes")
        self._tmp_path = os.path.join(self._pred_genes_dir, "tmp")
        self.cpus = 1
        self.create_dirs([self._pred_genes_dir, self._tmp_path])

        # Get genetic_code from dataset.cfg file
        # bacteria/archaea=11; Entomoplasmatales,Mycoplasmatales=4
        try:
            self._genetic_code = self.config.get(
                "prodigal", "prodigal_genetic_code"
            ).split(",")
        except NoOptionError:
            self._genetic_code = ["11"]

        # Set the ambiguous coding density range
        try:
            self._cd_upper = (
                float(self.config.get("prodigal", "ambiguous_cd_range_upper"))
                if len(self._genetic_code) > 1
                else 0
            )
        except NoOptionError:
            raise SystemExit(
                "Dataset config file does not contain required information. Please upgrade datasets."
            )

        self.current_gc = None
        self._current_run_mode = None
        self._tmp_name = None

        self.output_faa = os.path.join(self._pred_genes_dir, "predicted.faa")
        self._output_fna = os.path.join(self._pred_genes_dir, "predicted.fna")
        self.sequences_aa = {}
        self.sequences_nt = {}
        self.gene_details = {}

        self._input_length = self._get_genome_length()
        self._run_mode = ["single", "meta"] if self._input_length > 100000 else ["meta"]

        self.init_checkpoint_file()
        self.run_number += 1

    @property
    def output_folder(self):
        return self._output_folder

    def check_tool_dependencies(self):
        pass

    def generate_job_args(self):
        yield

    @log(
        "Genetic code {} selected as optimal",
        logger,
        attr_name="current_gc",
        on_func_exit=True,
    )
    def run(self):
        """
        1) If genome length > 100000 run in "single" mode, then "meta" mode if there are no gene predictions. Otherwise
        just run in "meta" mode. This is based on the recommendations in the Prodigal docs.
        2) Run once using genetic code 11. This can be overridden if the user includes a spceific genetic code in the
        config file.
        3) Check the genome coding density. If the coding density is above the ambiguous range (typically 0.73-0.8)
        then continue with the current genetic code. The ambiguous range was determined based on analysis done by Mose
        Manni. Otherwise run again on the next genetic code specified.
        4) If the next run still has a genetic density within the ambiguous range, read the stdout log files (formerly
        the GFF files) and extract the scores assigned to each gene prediction. Whichever genetic code yields the
        greatest mean score is selected.
        :return:
        """
        super().run()
        tmp_files = []

        for ix, m in enumerate(self._run_mode):
            self._current_run_mode = m
            for g in self._genetic_code:
                self.current_gc = g

                file_id = os.path.join(
                    self._tmp_path,
                    "prodigal_mode_{0}_code_{1}".format(
                        self._current_run_mode, self.current_gc
                    ),
                )
                self._tmp_name = "{}.faa".format(file_id)
                self.logfile_path_out = "{}_out.log".format(file_id)
                self.logfile_path_err = "err".join(
                    self.logfile_path_out.rsplit("out", 1)
                )  # Replace only the last occurence of "out" substring
                self._gc_run_results[self.current_gc].update(
                    {"tmp_name": self._tmp_name, "log_file": self.logfile_path_out}
                )

                if os.path.exists(
                    self._tmp_name
                ):  # Check to see if has already been run with these parameters
                    coding_density = self._gc_run_results[g]["cd"]
                else:
                    logger.info(
                        "Running Prodigal with genetic code {} in {} mode".format(
                            self.current_gc, self._current_run_mode
                        )
                    )
                    self.total = 1
                    self.run_jobs()
                    coding_length = self._get_coding_length(self.logfile_path_out)
                    coding_density = coding_length / self._input_length
                    self._gc_run_results[self.current_gc].update({"cd": coding_density})

                logger.debug("Coding density is {}".format(coding_density))
                tmp_files.append(self._gc_run_results[self.current_gc]["tmp_name"])

                # if the coding density is above the ambiguous range, then continue with these parameters
                if coding_density >= self._cd_upper:
                    break

            # If output files from both runs in "single" mode are empty, run again in "meta" mode, else raise Exception.
            if not any([os.stat(tmp_file).st_size > 0 for tmp_file in tmp_files]):
                if ix + 1 == len(self._run_mode):
                    raise NoGenesError("Prodigal")
                else:
                    continue

            # if only one genetic code to consider, proceed with it
            # if there is more than one possible set of parameters, decide which to use
            self.current_gc = (
                self._select_best_gc() if len(tmp_files) > 1 else self._genetic_code[0]
            )

            selected_logfile = self._gc_run_results[self.current_gc]["log_file"]
            selected_tmpfile = self._gc_run_results[self.current_gc]["tmp_name"]

            self._organize_prodigal_files(selected_tmpfile, selected_logfile)
            self.get_gene_details()
            self._gc_run_results[self.current_gc].update(
                {
                    "seqs_aa": self.sequences_aa,
                    "seqs_nt": self.sequences_nt,
                    "gene_details": self.gene_details,
                }
            )
            break

        return

    def configure_job(self, *args):
        tmp_name_nt = self._tmp_name.rpartition(".faa")[0] + ".fna"

        prodigal_job = self.create_job()
        prodigal_job.add_parameter("-p")
        prodigal_job.add_parameter("%s" % self._current_run_mode)
        prodigal_job.add_parameter("-f")
        prodigal_job.add_parameter("gff")
        prodigal_job.add_parameter("-g")
        prodigal_job.add_parameter("%s" % self.current_gc)
        prodigal_job.add_parameter("-a")
        prodigal_job.add_parameter("%s" % self._tmp_name)
        prodigal_job.add_parameter("-d")
        prodigal_job.add_parameter("%s" % tmp_name_nt)
        prodigal_job.add_parameter("-i")
        prodigal_job.add_parameter("%s" % self.input_file)

        return prodigal_job

    def get_gene_details(self):
        self.gene_details = defaultdict(list)

        with open(self._output_fna, "rU") as f:
            for record in SeqIO.parse(f, "fasta"):
                gene_name = record.id
                self.sequences_nt[gene_name] = record
                description_parts = record.description.split()
                gene_start = int(description_parts[2])
                gene_end = int(description_parts[4])
                strand = "+" if int(description_parts[6]) == 1 else "-"
                self.gene_details[gene_name].append(
                    {"gene_start": gene_start, "gene_end": gene_end, "strand": strand}
                )

        with open(self.output_faa, "rU") as f:
            for record in SeqIO.parse(f, "fasta"):
                self.sequences_aa[record.id] = record

        return

    @staticmethod
    def _get_coding_length(out_logfile):
        total_coding_length = 0
        with open(out_logfile, "r") as f:
            for line in f:
                if not line.startswith("#"):
                    try:
                        start = int(line.split("\t")[3])
                        stop = int(line.split("\t")[4])
                        total_coding_length += stop - start
                    except IndexError:
                        continue
                    except ValueError:
                        continue
        return total_coding_length

    def _get_genome_length(self):
        length_seqs = 0
        for line in open(self.input_file):
            if not line.startswith(">"):
                length_seqs += len(line)
        return length_seqs

    def _get_mean_score(self, gc):
        logfile = self._gc_run_results[gc]["log_file"]
        scores = []
        with open(logfile, "r") as f:
            for line in f:
                try:
                    score = re.search(";score=(.+?);", line).group(1)
                    scores.append(float(score))
                except AttributeError:
                    continue
        mean_score = sum(scores) / len(scores)
        return mean_score

    def _organize_prodigal_files(self, tmp_file, tmp_logfile):

        shutil.copy(tmp_file, self.output_faa)
        shutil.copy(tmp_file.rpartition(".faa")[0] + ".fna", self._output_fna)

        # copy selected log files from tmp/ to logs/
        new_logname = os.path.join(self.log_folder, "prodigal_out.log")
        shutil.copy(tmp_logfile, new_logname)
        shutil.copy(
            tmp_logfile.rpartition("_out.log")[0] + "_err.log",
            new_logname.rpartition("_out.log")[0] + "_err.log",
        )
        return

    def _select_best_gc(self):
        gcs, cds = zip(
            *[[gc, self._gc_run_results[gc]["cd"]] for gc in self._genetic_code]
        )
        sort_order = np.argsort(np.array(cds))[::-1]
        gcs_sorted = np.array(gcs)[sort_order]
        cds_sorted = np.array(cds)[sort_order]
        if abs(cds_sorted[0] - cds_sorted[1]) <= 0.05:
            mean_score1 = self._get_mean_score(gcs_sorted[0])
            mean_score2 = self._get_mean_score(gcs_sorted[1])
            gc = gcs_sorted[int(mean_score2 > mean_score1)]
        else:
            gc = gcs_sorted[0]
        return gc

    def get_version(self):
        prodigal_version = subprocess.check_output(
            [self.cmd, "-v"], stderr=subprocess.STDOUT, shell=False
        )
        prodigal_version = prodigal_version.decode("utf-8")
        prodigal_version = prodigal_version.split("\n")[1].split(":")[0]
        prodigal_version = prodigal_version.replace("Prodigal V", "")
        return prodigal_version



#!/usr/bin/env python3
# coding: utf-8
"""
.. versionadded:: 3.0.0
.. versionchanged:: 5.0.0

Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

This script proceeds to the BUSCO packages installation

"""

from distutils.core import setup

version = {}
with open("src/busco/_version.py") as version_file:
    exec(version_file.read(), version)

setup(
    name="BUSCO",
    version=version["__version__"],
    author="ezlab",
    license="Licensed under the MIT license. See LICENSE.md file.",
    author_email="ez@ezlab.org",
    long_description="Assessing genome assembly and annotation completeness "
    "with Benchmarking Universal Single-Copy Orthologs ",
    url="https://busco.ezlab.org/",
    platforms="Unix like",
    packages=["busco", "busco.analysis", "busco.busco_tools"],
    package_dir={"busco": "src/busco"},
    scripts=["bin/busco"],
)



import unittest
from unittest.mock import patch, call, Mock
from busco import ConfigManager
import os
import importlib


class TestConfigManager(unittest.TestCase):
    def setUp(self):
        self.base_config = "config/config.ini"
        self.params = {"config_file": "path/to/config.ini"}

        # In order to suppress logs, we need to patch the LogDecorator object that handles the log messages and the
        # loggers from each module. The following code was adapted from a StackOverflow answer here:
        # https://stackoverflow.com/a/37890916/4844311

        # Do cleanup first so it is ready if an exception is raised
        def kill_patches():  # Create a cleanup callback that undoes our patches
            patch.stopall()  # Stops all patches started with start()
            importlib.reload(
                ConfigManager
            )  # Reload our UUT module which restores the original decorator

        self.addCleanup(
            kill_patches
        )  # We want to make sure this is run so we do this in addCleanup instead of tearDown

        # Now patch the decorator where the decorator is being imported from
        patch(
            "busco.BuscoLogger.LogDecorator", lambda *args, **kwargs: lambda f: f
        ).start()  # The lambda makes our decorator into a pass-thru. Also, don't forget to call start()
        # HINT: if you're patching a decor with params use something like:
        # lambda *x, **y: lambda f: f
        importlib.reload(
            ConfigManager
        )  # Reloads the module which applies our patched decorator

    def test_get_config_from_params(self):

        config_manager = ConfigManager.BuscoConfigManager(self.params)
        self.assertEqual(config_manager.config_file, self.params["config_file"])

    def test_get_config_from_env(self):
        os.environ["BUSCO_CONFIG_FILE"] = "path/to/config.ini"
        with patch("os.access") as mockaccess:
            mockaccess.return_value = True
            config_manager = ConfigManager.BuscoConfigManager({})
            self.assertEqual(
                config_manager.config_file, os.environ.get("BUSCO_CONFIG_FILE")
            )

    @patch("__main__.ConfigManager_unittests.ConfigManager.logger", autospec=True)
    def test_config_validated(self, *args):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        with patch(
            "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain",
            autospec=True,
        ):
            config_manager.load_busco_config()
            config_manager.config.validate.assert_called()

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_log_warning_if_neither_lineage_nor_autolineage_specified(
        self, mock_config_main, *args
    ):
        mock_config_main.return_value.check_lineage_present = lambda: False
        mock_config_main.return_value.getboolean.side_effect = [
            False,
            False,
            True,
            False,
            True,
        ]  # These values comprise the three distinct logical paths into this part
        # of the code: run1: False, False; run2: True, (shortcircuit); run3: False, True.
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            with self.assertLogs(ConfigManager.logger, "WARNING"):
                config_manager.load_busco_config()
            with patch("busco.ConfigManager.logger.warning") as mock_logger:
                for _ in range(2):
                    config_manager.load_busco_config()
                    self.assertFalse(mock_logger.called)

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_log_warning_if_both_lineage_and_autolineage_specified(
        self, mock_config_main, *args
    ):
        mock_config_main.return_value.check_lineage_present = lambda: True
        mock_config_main.return_value.getboolean.side_effect = [
            True,
            False,
            True,
            False,
            False,
        ]
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            for _ in range(2):
                with self.assertLogs(ConfigManager.logger, "WARNING"):
                    config_manager.load_busco_config()
            with patch("busco.ConfigManager.logger.warning") as mock_logger:
                config_manager.load_busco_config()
                self.assertFalse(mock_logger.called)

    patch("busco.ConfigManager.logger.warning")

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_config_updated_if_lineage_missing(self, mock_config_main, *args):
        mock_config_main.return_value.check_lineage_present = lambda: False
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            config_manager.load_busco_config()
        calls = [
            call("busco_run", "auto-lineage", "True"),
            call("busco_run", "lineage_dataset", test_dataset_path),
        ]
        mock_config_main.return_value.set.assert_has_calls(calls, any_order=True)

    @patch("busco.ConfigManager.logger.warning")
    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_config_updated_if_lineage_present(self, mock_config_main, *args):
        mock_config_main.return_value.check_lineage_present = lambda: True
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            config_manager.load_busco_config()
        calls = [
            call("busco_run", "auto-lineage", "False"),
            call("busco_run", "auto-lineage-prok", "False"),
            call("busco_run", "auto-lineage-euk", "False"),
        ]
        mock_config_main.return_value.set.assert_has_calls(calls, any_order=True)

    @patch("busco.ConfigManager.logger.warning")
    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_update_dirname(self, mock_config_main, *args):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        mock_config_main.return_value.get = lambda *args: test_dataset_path
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            config_manager.load_busco_config()
        mock_config_main.return_value.set_results_dirname.assert_called_with(
            test_dataset_path
        )

    @patch("busco.ConfigManager.logger.warning")
    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_lineage_downloaded(self, mock_config_main, *args):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        mock_config_main.return_value.get = lambda *args: test_dataset_path
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            config_manager.load_busco_config()
        mock_config_main.return_value.download_lineage_file.assert_called_with(
            test_dataset_path
        )

    @patch("busco.ConfigManager.logger.warning")
    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_lineage_dataset_config_loaded(self, mock_config_main, *args):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", lambda: test_dataset_path
        ):
            config_manager.load_busco_config()
        mock_config_main.return_value.load_dataset_config.assert_called()

    @patch("busco.ConfigManager.logger.warning")
    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.BuscoConfigMain", autospec=True
    )
    def test_run_auto_select_if_no_lineage(self, mock_config_main, *args):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        mock_config_main.return_value.check_lineage_present = lambda: False
        test_dataset_path = "/path/to/lineage_dataset"
        with patch.object(
            config_manager, "auto_select_lineage", return_value=test_dataset_path
        ):
            config_manager.load_busco_config()
            config_manager.auto_select_lineage.assert_called()

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.AutoSelectLineage",
        autospec=True,
    )
    def test_auto_select_lineage_call_function_initializes_asl(self, mock_asl):
        mock_asl.return_value.best_match_lineage_dataset = Mock()
        mock_asl.return_value.selected_runner = Mock()
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        config_manager.auto_select_lineage()
        mock_asl.assert_called()

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.AutoSelectLineage",
        autospec=True,
    )
    def test_auto_select_lineage_call_function_runs_asl(self, mock_asl):
        mock_asl.return_value.best_match_lineage_dataset = Mock()
        mock_asl.return_value.selected_runner = Mock()
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        config_manager.auto_select_lineage()
        mock_asl.return_value.run_auto_selector.assert_called()

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.AutoSelectLineage",
        autospec=True,
    )
    def test_auto_select_lineage_call_function_gets_lineage_dataset(self, mock_asl):
        mock_asl.return_value.best_match_lineage_dataset = Mock()
        mock_asl.return_value.selected_runner = Mock()
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        config_manager.auto_select_lineage()
        mock_asl.return_value.get_lineage_dataset.assert_called()

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.AutoSelectLineage",
        autospec=True,
    )
    def test_auto_select_lineage_call_function_returns_lineage_dataset(self, mock_asl):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        lineage_dataset = "best_match_dataset"
        mock_asl.return_value.best_match_lineage_dataset = lineage_dataset
        mock_asl.return_value.selected_runner = Mock()
        retval = config_manager.auto_select_lineage()
        self.assertEqual(retval, lineage_dataset)

    @patch(
        "__main__.ConfigManager_unittests.ConfigManager.AutoSelectLineage",
        autospec=True,
    )
    def test_auto_select_lineage_call_function_selects_runner(self, mock_asl):
        config_manager = ConfigManager.BuscoConfigManager(self.params)
        mock_asl.return_value.best_match_lineage_dataset = Mock()
        mock_asl.return_value.selected_runner = "test"
        config_manager.auto_select_lineage()
        self.assertEqual("test", config_manager.runner)

    def tearDown(self):
        pass



import unittest
from busco.analysis import GenomeAnalysis
from unittest.mock import patch, Mock


class TestConfigManager(unittest.TestCase):
    def setUp(self) -> None:
        pass

    # @patch('busco.analysis.GenomeAnalysis.BuscoAnalysis.config.get', return_value="test")
    # @patch('busco.analysis.GenomeAnalysis.BuscoAnalysis.config.getboolean', return_value=True)
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config")
    @patch("busco.analysis.BuscoAnalysis.os.path")
    @patch("busco.analysis.GenomeAnalysis.NucleotideAnalysis.check_nucleotide_file")
    def test_init_eukaryota_augustus_checks_filetype(self, mock_check_nucl_file, *args):
        GenomeAnalysis.GenomeAnalysisEukaryotesAugustus()
        mock_check_nucl_file.assert_called()

    @patch("busco.analysis.GenomeAnalysis.NucleotideAnalysis.__init__")
    # @patch('busco.analysis.GenomeAnalysis.BuscoAnalysis.config.get')
    # @patch('busco.analysis.GenomeAnalysis.BuscoAnalysis.config.getboolean')
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config")
    @patch("busco.analysis.Analysis.logger.warning")
    @patch("busco.analysis.GenomeAnalysis.OptimizeAugustusRunner")
    @patch("busco.analysis.GenomeAnalysis.ETrainingRunner")
    @patch("busco.analysis.GenomeAnalysis.NewSpeciesRunner")
    @patch("busco.analysis.GenomeAnalysis.GFF2GBRunner")
    @patch("busco.analysis.GenomeAnalysis.AugustusRunner")
    @patch("busco.analysis.Analysis.TBLASTNRunner")
    @patch("busco.analysis.Analysis.MKBLASTRunner")
    @patch("busco.analysis.BuscoAnalysis.HMMERRunner")
    def test_init_tools_eukaryota_augustus(
        self,
        mock_hmmer,
        mock_mkblast,
        mock_tblastn,
        mock_augustus,
        mock_gff2gb,
        mock_new_species,
        mock_etraining,
        mock_optimize_augustus,
        *args
    ):
        analysis = GenomeAnalysis.GenomeAnalysisEukaryotesAugustus()
        analysis.init_tools()
        mock_hmmer.assert_called()
        mock_mkblast.assert_called()
        mock_tblastn.assert_called()
        mock_augustus.assert_called()
        mock_gff2gb.assert_called()
        mock_new_species.assert_called()
        mock_etraining.assert_called()
        mock_optimize_augustus.assert_called()

    @patch("busco.analysis.GenomeAnalysis.NucleotideAnalysis.__init__")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config.get")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config", autospec=True)
    @patch("busco.analysis.GenomeAnalysis.MetaeukRunner")
    @patch("busco.analysis.BuscoAnalysis.HMMERRunner")
    def test_init_tools_eukaryota_metaeuk(self, mock_hmmer, mock_metaeuk, *args):
        analysis = GenomeAnalysis.GenomeAnalysisEukaryotesMetaeuk()
        analysis.init_tools()
        mock_hmmer.assert_called()
        mock_metaeuk.assert_called()

    @patch("busco.analysis.GenomeAnalysis.NucleotideAnalysis.__init__")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config.get")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config", autospec=True)
    @patch("busco.analysis.GenomeAnalysis.ProdigalRunner")
    @patch("busco.analysis.BuscoAnalysis.HMMERRunner")
    def test_init_tools_prokaryota(self, mock_hmmer, mock_prodigal, *args):
        analysis = GenomeAnalysis.GenomeAnalysisProkaryotes()
        analysis.init_tools()
        mock_hmmer.assert_called()
        mock_prodigal.assert_called()

    @patch("busco.analysis.GenomeAnalysis.NucleotideAnalysis.__init__")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config.get")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.config", autospec=True)
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.run_analysis")
    @patch("busco.analysis.GenomeAnalysis.BuscoAnalysis.run_hmmer")
    @patch("busco.analysis.GenomeAnalysis.GenomeAnalysisEukaryotesMetaeuk._run_metaeuk")
    def test_run_analysis_metaeuk(self, mock_run_metaeuk, mock_run_hmmer, *args):
        analysis = GenomeAnalysis.GenomeAnalysisEukaryotesMetaeuk()
        analysis.metaeuk_runner = Mock()
        analysis.hmmer_runner = Mock(missing_buscos=[])
        analysis.hmmer_runner.fragmented_buscos.keys = Mock(return_value=[])
        analysis.gene_details = Mock(autospec=True)
        analysis.sequences_aa = Mock(autospec=True)
        analysis.sequences_nt = Mock(autospec=True)
        analysis.run_analysis()
        mock_run_metaeuk.assert_called()
        mock_run_hmmer.assert_called()

        # @patch('busco.GenomeAnalysis.GenomeAnalysisEukaryotesAugustus._rerun_analysis')
        # @patch('busco.GenomeAnalysis.GenomeAnalysisEukaryotesAugustus.run_hmmer')
        # @patch('busco.GenomeAnalysis.GenomeAnalysisEukaryotesAugustus._run_augustus')
        # @patch('busco.GenomeAnalysis.BLASTAnalysis._run_tblastn')
        # @patch('busco.GenomeAnalysis.BLASTAnalysis._run_mkblast')
        # mock_mkblast.assert_called()
        # mock_tblastn.assert_called()
        # mock_augustus.assert_called()
        # mock_hmmer.assert_called()
        # mock_rerun.assert_called()

    def tearDown(self) -> None:
        pass



#!/usr/bin/env python3
# coding: utf-8
"""
.. package:: busco
   :synopsis: BUSCO - Benchmarking Universal Single-Copy Orthologs.


Copyright (c) 2016-2021, Evgeny Zdobnov (ez@ezlab.org)
Licensed under the MIT license. See LICENSE.md file.

"""
from ._version import __version__ as version

__all__ = [
    "Actions",
    "Analysis",
    "AutoLineage",
    "BuscoAnalysis",
    "BuscoConfig",
    "BuscoDownloadManager",
    "BuscoLogger",
    "BuscoPlacer",
    "BuscoRunner",
    "BuscoTools",
    "ConfigManager",
    "GeneSetAnalysis",
    "GenomeAnalysis",
    "Toolset",
    "TranscriptomeAnalysis",
    "BuscoConfig",
    "BuscoPlacer",
]
__version__ = version
